#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include <ncurses.h>
#include <unistd.h>
#include <termios.h>
#include <limits.h>
#include <time.h>
#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>

#include "hash.h"
#include "structure.h"
#include "base.h"
#include "move.h"
#include "display.h"
#include "generation.h"

int main(int argc, char** argv) {
    srand(time(NULL));

    //-------------------------------------------------------------------------------//
    
    if (SDL_Init(SDL_INIT_EVERYTHING) != 0) {
        printf("error initializing SDL: %s\n", SDL_GetError());
    }
    SDL_Window* win = SDL_CreateWindow("Game",
                                       SDL_WINDOWPOS_CENTERED,
                                       SDL_WINDOWPOS_CENTERED,
                                       1500, 1000, 0);
 
    Uint32 render_flags = SDL_RENDERER_ACCELERATED;
    SDL_Renderer* rend = SDL_CreateRenderer(win, -1, render_flags);
    SDL_SetRenderDrawBlendMode(rend, SDL_BLENDMODE_BLEND);
    
    //-------------------------------------------------------------------------------//

    if(SDL_Init(SDL_INIT_AUDIO)) {
        fprintf(stderr, "cannot initialize audio");
        exit(1);
    }
    SDL_SetRelativeMouseMode(true) ;

    /* -------------------------------------------------------- */
    init_csts() ;
    init_hashtbl() ;
    init_draworder() ;
    import_digits(rend) ;
    import_letters(rend) ;
    while(true) {
        resetRenderer(rend) ;

        SDL_SetRenderDrawColor(rend, 255, 255, 255, SDL_ALPHA_OPAQUE) ;
        playerActions() ;
        generate_nearby_chunks(1);
        drawCurrentRoom(rend);
        drawData(rend) ;

        updateRenderer(rend) ;
        usleep(1000000/60) ;
    }
    free_digits(digits) ;

    /* -------------------------------------------------------- */

    SDL_DestroyRenderer(rend);
    SDL_DestroyWindow(win);
    SDL_Quit();

    /* -------------------------------------------------------- */

    printf("Done\n") ;

    return 0;
}