#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include <ncurses.h>
#include <unistd.h>
#include <termios.h>
#include <limits.h>
#include <time.h>
#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>

#include "hash.h"
#include "structure.h"
#include "base.h"
#include "move.h"

// ---------------------------------------------------------------------------------------------------- //
double sensitivity = 0.22 ;
double fov = 90.0 ;
double speed = 1.0 ;
double min_dist = 0.2 ;
// ---------------------------------------------------------------------------------------------------- //

double camx ;
double camy ;
double camz ;

double rot_hz ;
double rot_vt ;

double tan_fov ;

bool has_changed ;

void init_csts() {
    camx = 3.0 ;
    camy = 3.0 ;
    camz = 3.0 ;
    rot_hz = 0.0 ;
    rot_vt = 180.0 ;
    tan_fov = tan((fov * 3.14159 / 180.0) / 2.0) ;
}

bool is_colliding() {
    for(int k = 0; k < current_room->map_size; k++) {
        double dist = distance_pt_cube_0_3d(camx, camy, camz, current_room->map[k]) ;
        if(dist <= min_dist) {
            return true ;
        }
    }
    for(int k = 0; k < current_room->tps_size; k++) {
        double dist = distance_pt_cube_0_3d(camx, camy, camz, current_room->tps[k].hitbox) ;
        if(dist <= min_dist) {
            player_chx = current_room->tps[k].dest_chx ;
            player_chy = current_room->tps[k].dest_chy ;
            camx = current_room->tps[k].dest_x ;
            camy = current_room->tps[k].dest_y ;
            camz = current_room->tps[k].dest_z ;
            current_room = hashtbl_find_opt(visited, player_chx, player_chy);
            return true ;
        }
    }
    return false ;
}

void playerActions() {
    SDL_Event event;
    while(SDL_PollEvent(&event)) {
        switch (event.type) {
            case SDL_QUIT:
                break;
            case SDL_MOUSEMOTION:
                has_changed = true ;
                rot_hz += sensitivity * event.motion.xrel / 100.0 ;
                rot_vt -= sensitivity * event.motion.yrel / 100.0 ;
                if(rot_hz >= 2*3.141592) {
                    rot_hz -= 2*3.141592 ;
                } else if(rot_hz < 0.0) {
                    rot_hz += 2*3.141592 ;
                }
                if(rot_vt <= 3.14159/2.0) {
                    rot_vt = 3.14159/2.0;
                } else if(rot_vt >= 3.0*3.14159/2.0) {
                    rot_vt = 3.0*3.14159/2.0 ;
                }
                break;
            case SDL_KEYDOWN:
                has_changed = true ;
                switch (event.key.keysym.sym) {
                    case SDLK_z:
                        for(int k = 0; k < 10; k++) {
                            camz += speed*cos(rot_hz)*cos(rot_vt)/10;
                            camx -= speed*sin(rot_hz)*cos(rot_vt)/10;
                            camy += speed*sin(rot_vt)/10;
                            if(is_colliding()) {
                                camz -= speed*cos(rot_hz)*cos(rot_vt)/10;
                                camx += speed*sin(rot_hz)*cos(rot_vt)/10;
                                camy -= speed*sin(rot_vt)/10;
                                k=11;
                            }
                        }
                        break;
                    case SDLK_q:
                        for(int k = 0; k < 10; k++) {
                            camx -= speed*cos(rot_hz)/10;
                            camz -= speed*sin(rot_hz)/10;
                            if(is_colliding()) {
                                camx += speed*cos(rot_hz)/10;
                                camz += speed*sin(rot_hz)/10;
                                k=11;
                            }
                        }
                        break;
                    case SDLK_s:
                        for(int k = 0; k < 10; k++) {
                            camz -= speed*cos(rot_hz)*cos(rot_vt)/10;
                            camx += speed*sin(rot_hz)*cos(rot_vt)/10;
                            camy -= speed*sin(rot_vt)/10;
                            if(is_colliding()) {
                                camz += speed*cos(rot_hz)*cos(rot_vt)/10;
                                camx -= speed*sin(rot_hz)*cos(rot_vt)/10;
                                camy += speed*sin(rot_vt)/10;
                                k=11;
                            }
                        }
                        break;
                    case SDLK_d:
                        for(int k = 0; k < 10; k++) {
                            camx += speed*cos(rot_hz)/10;
                            camz += speed*sin(rot_hz)/10;
                            if(is_colliding()) {
                                camx -= speed*cos(rot_hz)/10;
                                camz -= speed*sin(rot_hz)/10;
                                k=11;
                            }
                        }
                        break;

                    case SDLK_t:
                        fprintf(stderr, "Killed.\n") ;
                        exit(1) ;
                        break;
                    
                    case SDLK_a:
                        rot_hz -= sensitivity ;
                        break;
                    case SDLK_e:
                        rot_hz += sensitivity ;
                        break;
                    
                    case SDLK_p:
                        rot_vt -= sensitivity ;
                        break;
                    case SDLK_m:
                        rot_vt += sensitivity ;
                        break;
                    
                    default:
                        break;
                }
        }
    }
}